#ifndef __UTILS_H
#define __UTILS_H

#include <cstdio>
#include <cstdlib>

#include <string>
#include <sstream>
#include <fstream>
#include <algorithm>

#include <unistd.h>
#include <sys/time.h>


namespace utils {

#define STRX(x) #x
#define STR(x) STRX(x)

/******************************* SYSTEM CALLS *********************************/

void execute(std::string cmd) {
  int system_ret = system(cmd.c_str());
  if (system_ret) {
    fprintf(stderr, "Error: executing command [%s] returned %d.\n",
        cmd.c_str(), system_ret);
    exit(1);
  }
}

/****************************** MEASURING TIME ********************************/

long double wclock() {
  timeval tim;
  gettimeofday(&tim, NULL);

  return tim.tv_sec + (tim.tv_usec / 1000000.0L);
}

/**************************** FILE MANIPULATION *******************************/

// Basic routines.
FILE *open_file(std::string fname, std::string mode) {
  FILE *f = fopen(fname.c_str(), mode.c_str());
  if (!f) {
    perror(fname.c_str());
    std::exit(EXIT_FAILURE);
  }

  return f;
}

long file_size(std::string fname) {
  FILE *f = open_file(fname, "rt");
  std::fseek(f, 0L, SEEK_END);
  long result = ftell(f);
  std::fclose(f);

  return result;
}

bool file_exists(std::string fname) {
  FILE *f = std::fopen(fname.c_str(), "r");
  bool result = (f != NULL);
  if (f != NULL)
    std::fclose(f);

  return result;
}

// Writing to file.
template<typename T>
void add_objects_to_file(T *tab, size_t length, FILE *f) {
  size_t fwrite_ret = fwrite(tab, sizeof(T), length, f);
  if (fwrite_ret != length) {
    fprintf(stderr, "Error: fwrite in line %s of %s returned %ld.\n",
        STR(__LINE__), STR(__FILE__), (long)fwrite_ret);
    std::exit(EXIT_FAILURE);
  }
}

// Reading from file.
template<typename T>
void read_file(T* &tab, long &length, std::string filename) {
  FILE *f = open_file(filename, "r");

  std::fseek(f, 0L, SEEK_END);
  length = ftell(f) / sizeof(T);
  if (length == 0) {
    fprintf(stderr, "Error: reading empty file.\n");
    std::exit(EXIT_FAILURE);
  }

  std::rewind(f);
  tab = new T[length];

  size_t fread_ret = std::fread(tab, sizeof(T), length, f);
  if ((long)fread_ret != length) {
    fprintf(stderr, "Error: fread in line %s of %s returned %lu\n",
        STR(__LINE__), STR(__FILE__), fread_ret);
    std::exit(EXIT_FAILURE);
  }

  std::fclose(f);
}

/********************************* MISC ***************************************/

void find_stxxl_config() {
  if (file_exists("./.stxxl")) {
    fprintf(stderr, "STXXL config file detected.\n");

    return;
  } else if (file_exists(std::string(getenv("HOME")) + "/.stxxl")) {
    fprintf(stderr, "Cannot find STXXL config file. Using $HOME/.stxxl\n");
    execute("cp " + std::string(getenv("HOME")) + "/.stxxl ./");

    return;
  } else {  
    fprintf(stderr, "Error: failed to find/copy STXXL config file!\n");
    std::exit(EXIT_FAILURE);
  }
}

} // namespace utils

#endif // __UTILS_H
